// ijvsparse.cpp

#include "ijvsparse.h"

/*-----------------------------------------------------!
! Conjugate Gradient Method for a Sparse Linear System !
!------------------------------------------------------!
! Ref.: "Numerical Recipes, Cambridge University Press !
!        1986 (chapter 2.10)" [BIBLI 08].              !
!-----------------------------------------------------*/

void ax(int n, int na, int *ia, int *ja, double *va, double *x, double *y) {
    int i;
    for(i=0; i<n; i++) y[i]=0.0;
    for(i=0; i<na; i++) y[ia[i]]=y[ia[i]]+va[i]*x[ja[i]];   // y = A.x
}

void atx(int n, int na, int *ia, int *ja, double *va, double *x, double *y) {
    int i;
    for(i=0; i<n; i++) y[i]=0.0;
    for(i=0; i<na; i++) y[ja[i]]=y[ja[i]]+va[i]*x[ia[i]];   // y = At.x
}

void ijvsolve(int n, int na, int *ia, int *ja, double *va, double *b, double *x, double *rsq)  {
/*--------------------------------------------------------------------!
! Solves the linear system A.x = b for the vector x of length n,      !
! given the right-hand vector b, and given two subroutines,           !
! ax(x,y) and atx(x,y), which respectively calculate A.x and At.x     !
! (At for Transpose of A) for x given as first argument,              !
! returning the result in their second argument: y.                   !
! These subroutines should take every advantage of the sparseness     !
! of the matrix A. On input, x should be set to a first guess of the  !
! desired solution (all zero components is fine). On output, x is     !
! the solution vector, and rsq is the sum of the squares of the       !
! components of the residual vector A.x - b. If this is not small,    !
! then the matrix is numerically singular and the solution represents !
! a least-squares approximation.                                      !
!--------------------------------------------------------------------*/
// LABEL: L1, E
   double *g, *h, *xi, *xj;
   double eps,eps2,bsq,rp,anum,aden,gam,gg,dgg;
   int i,irst,iter;
   g = new double[n];
   h = new double[n];
   xi = new double[n];
   xj = new double[n];
   for(i=0; i<n; i++) g[i]=h[i]=xi[i]=xj[i]=0.0;
   eps=1.0e-6;
   eps2=eps*eps*(double)n;
   irst=0;
L1:irst++;
   ax(n,na,ia,ja,va,x,xi);
   rp=0.0;
   bsq=0.0;
   for (i=0; i<n; i++) {
     bsq=bsq+b[i]*b[i];
     xi[i]=xi[i]-b[i];
     rp=rp+xi[i]*xi[i];
   }
   atx(n,na,ia,ja,va,xi,g);
   for (i=0; i<n; i++) {
     g[i]=-g[i];
     h[i]=g[i];
   }
   for (iter=0; iter<5*n; iter++) {  //main loop
     ax(n,na,ia,ja,va,h,xi);
     anum=0.0;
     aden=0.0;
     for (i=0; i<n; i++) {
       anum=anum+g[i]*h[i];
       aden=aden+xi[i]*xi[i];
     }
     if (aden==0.0) quitter("Pb dans ijasolve() :\nVery singular matrix.\n\n");
     anum=anum/aden;
     for (i=0; i<n; i++) {
       xi[i]=x[i];
       x[i]=x[i]+anum*h[i];
     }
     ax(n,na,ia,ja,va,x,xj);
     *rsq=0.0;
     for (i=0; i<n; i++) {
       xj[i]=xj[i]-b[i];
       *rsq=*rsq+xj[i]*xj[i];
     }
     if (*rsq==rp || *rsq<=bsq*eps2) {
       goto E;  //normal return
     }
     if (*rsq > rp) {
       for (i=0; i<n; i++)  x[i]=xi[i];
       if (irst>=3) goto E;  //return if too many restarts
       goto L1;
     }
     rp=*rsq;
     atx(n,na,ia,ja,va,xj,xi);  //compute gradient for next iteration
     gg=0.0;
     dgg=0.0;
     for (i=0; i<n; i++) {
       gg=gg+g[i]*g[i];
       dgg=dgg+(xi[i]+g[i])*xi[i];
     }
     if (gg==0.0) {
       goto E;  //rare but normal return
     }
     gam=dgg/gg;
     for (i=0; i<n; i++) {
       g[i]=-xi[i];
       h[i]=g[i]+gam*h[i];
     }
   } //main loop
   quitter("Pb dans ijasolve() :\nToo many iterations.\n\n");
E: delete [] g;
   delete [] h;
   delete [] xi;
   delete [] xj;
} 
